/*
 * XTL - the eXternalization Template Library
 * Copyright (C) 1998-2000 Jose' Orlando Pereira, Universidade do Minho
 *
 * jop@di.uminho.pt - http://gsd.di.uminho.pt/~jop
 *
 * Departamento de Informatica, Universidade do Minho
 * Campus de Gualtar, 4710-057 Braga, Portugal
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 * $Id: giop.h 1.12 Wed, 26 Apr 2000 18:14:08 +0100 jop $
 */

#ifndef __XTL_GIOP
#define __XTL_GIOP

#include <xtl/md.h>

template <class Buffer>
class GIOP_format: public generic_format<Buffer> {
 private:

	static const int lenWord = sizeof(void*);
	
#if   (__BYTE_ORDER == __LITTLE_ENDIAN)
	static const bool littleEndianMachine = true;
#elif (__BYTE_ORDER == __BIG_ENDIAN)
	static const bool littleEndianMachine = false;
#endif

	bool littleEndianBuffer;
	int align;

	// Add padding so that data (of size n) is aligned on
	// n byte boundaries
	int padding( int n ) {
		int pad = ((align % n) ? (n - (align % n)) : 0);
		align += n + pad;
		return pad;
	}

	char* req_align( int n ) {
		int pad = padding( n );
		return ( reinterpret_cast<char*>( require(n+pad) ) + pad );
	}

	char* des_align( int n ) {
		int pad = padding( n );
		return ( reinterpret_cast<char*>( desire(n+pad) ) + pad );
	}

	// Data is written to the buffer with the byte order of the local machine.
	// When reading a buffer (which may come from a different machine), the
	// ordering must be accounted for.
	void n2hs(char const in[2], short* out) {
		if( littleEndianBuffer == littleEndianMachine )
			*out=*reinterpret_cast<short const*>( in );
		else
			*out=bswap_16(*reinterpret_cast<unsigned short const*>( in ));
	}

	void n2hl(char const in[4], int* out) {
		if( littleEndianBuffer == littleEndianMachine )
			*out=*reinterpret_cast<int const*>( in );
		else
			*out=bswap_32(*reinterpret_cast<unsigned int const*>( in ));
	}

	void n2hh(char const in[8], longlong* out) {
		if( littleEndianBuffer == littleEndianMachine )
			*out=*reinterpret_cast<longlong const*>( in );
		else
			*out=bswap_64(*reinterpret_cast<unsignedlonglong const*>( in ));
	}

	inline void h2ns( short const* in, char out[2] )
		{ *reinterpret_cast<short*>( out ) = *in; }

	inline void h2nl( int const* in, char out[4] )
		{ *reinterpret_cast<int*>( out ) = *in; }

	inline void h2nh( longlong const* in, char out[8] )
		{ *reinterpret_cast<longlong*>( out ) = *in; }

 public:
 	typedef Buffer buffer;

	GIOP_format(Buffer& buf):
		generic_format<Buffer>(buf),
		littleEndianBuffer(littleEndianMachine),
		align(0) {}

	// Allows data output on one machine to be read on another with
	// (possibly) different byte ordering
	void input_start()
		{ input_simple(littleEndianBuffer); }
		
	void input_start_array(int& n)
		{ input_simple(n); } 
	bool input_end_array(int& n)
		{ return n--<=0; }

	void input_simple( bool& data )
		{ data=!!*req_align(1); }
 	void input_simple( char& data )
		{ data=*req_align(1); }
 	void input_simple( unsigned char& data )
		{ data=*req_align(1); }
 	void input_simple( short& data )
		{ n2hs(req_align(2), &data); }
 	void input_simple( unsigned short& data )
		{ n2hs(req_align(2), reinterpret_cast<short*>( &data )); }
 	void input_simple( int& data )
		{ n2hl(req_align(4), &data); }
 	void input_simple( unsigned int& data )
		{ n2hl(req_align(4), reinterpret_cast<int*>( &data )); }
	// long is stored as longlong to allow communication
	// between 32 and 64 bit machines
 	void input_simple( long& data )
		{ n2hh(req_align(8), reinterpret_cast<longlong*>( &data )); }
 	void input_simple( unsigned long& data )
		{ n2hh(req_align(8), reinterpret_cast<longlong*>( &data )); }
 	void input_simple( longlong& data )
		{ n2hh(req_align(8), &data); }
 	void input_simple( unsignedlonglong& data )
		{ n2hh(req_align(8), reinterpret_cast<longlong*>( &data )); }
 	void input_simple( float& data )
		{n2hl(req_align(4), reinterpret_cast<int*>( &data )); }
 	void input_simple(double& data)
		{n2hh(req_align(8), reinterpret_cast<longlong*>( &data )); }

	void input_chars(char* data, int size) {
		input_raw(data, size);
	}

	// This routine is identical to that in XDR_format
	void input_raw(char* data, int size) {
		int i;
		for(i=0;i<(size>>8)-1;i++,data+=256)
			memcpy(data, require(256), 256);
		int res=size-(i<<8);
		memcpy(data, require(res), res);
		if (res%lenWord!=0)
			require(lenWord-res%lenWord);
	}

//	void input_raw(char* data, int size) {
//		read(data, size);
//		align+=size;
//	}

 	void output_start() {
		align = 0;
		output_simple(littleEndianBuffer);
	}

	void output_start_array(int n) {output_simple(n);}
	void output_end_array() {}
	
 	void output_simple( bool const& data )
		{ *des_align(1)=data?1:0; }
 	void output_simple( char const& data )
		{ *des_align(1)=data; }
 	void output_simple( unsigned char const& data )
		{ *des_align(1)=data; }
 	void output_simple( short const& data )
		{ h2ns(&data, des_align(2)); }
 	void output_simple( unsigned short const& data )
		{ h2ns(reinterpret_cast<short const*>( &data ), des_align(2)); }
 	void output_simple( int const& data )
		{ h2nl(&data, des_align(4)); }
 	void output_simple( unsigned int const& data )
		{ h2nl(reinterpret_cast<int const*>( &data ), des_align(4)); }
 	void output_simple( long const& data )
		{ h2nh(reinterpret_cast<longlong const*>( &data ), des_align(8)); }
 	void output_simple( unsigned long const& data )
		{ h2nh(reinterpret_cast<longlong const*>( &data ), des_align(8)); }
 	void output_simple( longlong const& data )
		{ h2nh(&data, des_align(8)); }
 	void output_simple( unsignedlonglong const& data )
		{ h2nh(reinterpret_cast<longlong const*>( &data ), des_align(8)); }
 	void output_simple( float const& data )
		{ h2nl(reinterpret_cast<int const*>( &data ), des_align(4)); }
 	void output_simple( double const& data )
		{ h2nh(reinterpret_cast<longlong const*>( &data ), des_align(8)); }

	void output_chars(char const* data, int size) {
		output_raw(data, size);
	}

	// This routine is identical to that in XDR_format
	void output_raw(char const* data, int size) {
		int i;
		for(i=0;i<(size>>8)-1;i++,data+=256)
			memcpy(desire(256), data, 256);
		int res=size-(i<<8);
		memcpy(desire(res), data, res);
		if (res%lenWord!=0)
			memset(desire(lenWord-res%lenWord), 0, lenWord-res%lenWord);
	}

//	void output_raw(char const* data, int size) {
//		write(data, size);
//		align+=size;
//	}
};

#endif // __XTL_GIOP
